@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # coding: utf-8
 
-# In[ ]:
+# In[1]:
 
 
 get_ipython().run_line_magic('matplotlib', 'inline')
@@ -185,6 +185,45 @@ class fock_space:
                     return
                 n -= current_state[j+1] - 1
                 current_state = current_state[:j] + (current_state[j]+1, 0) + current_state[j+2:]
+                
+    def ngenerator(self, k, N1, q):
+        local_exc2 = np.delete(self.local_exc1, k)
+        current_state = (0,)*len(local_exc2)
+        n = 0
+        while True:
+            pp = list(current_state)
+            pp.insert(k,q)
+            pp = tuple(pp)
+            
+            yield self.find_index[pp]
+            
+            j = len(local_exc2) - 1
+            current_state = current_state[:j] + (current_state[j]+1,)
+            n += 1
+            while n > N1 or current_state[j] >= local_exc2[j]:
+                j -= 1
+                if j < 0:
+                    return
+                n -= current_state[j+1] - 1
+                current_state = current_state[:j] + (current_state[j]+1, 0) + current_state[j+2:]
+                
+    def ptrace(self, psi, k):
+        psi1 = coo_array(psi)
+        rho = psi1.conj().T @ psi1
+        
+        N = self.global_exc
+        result = np.zeros((N+1,N+1))
+        for p in range(N+1):
+            for q in range(N+1):
+                N1 = N - max(p,q)
+                a = list(self.ngenerator(k, N1, p))
+                b = list(self.ngenerator(k, N1, q))
+                for i in range(len(a)):
+                    result[p,q] += rho[a[i],b[i]]
+        return(result)
+                            
+            
+        
       
         
     def occupations(self,i):
@@ -274,7 +313,99 @@ class fock_space_kron:
                 self.create.append(sparse.kron(f1.create[k], f2.eye))
             for k in range(f1.modes, self.modes):
                 self.create.append(sparse.kron(f1.eye, f2.create[k-f1.modes]))
+                
+        self.local_exc = np.concatenate((f1.local_exc,f2.local_exc))
+        self.local_exc1 = np.array(self.local_exc+1) #internal parameter
+        self.global_exc = f1.global_exc + f2.global_exc
+        self.states_list = list(self.states_generator())
+        self.find_index = {state: idx for idx, state in enumerate(self.states_list)}
+                
+                
+                
+     #ggggggggggg           
+                
+    def states_generator(self):
+        #Generates all the possible states for given Fock space
+        current_state = (0,)*len(self.local_exc1)
+        n = 0
+        m = 0
+        j = len(self.local_exc1)
+        while True:
+            yield current_state
+
+            j = len(self.local_exc1) - 1
+            current_state = current_state[:j] + (current_state[j]+1,)
+            n += 1
             
+            while n > f2.global_exc or m > f1.global_exc or current_state[j] >= self.local_exc1[j]:
+                j -= 1
+                if j < 0:
+                    return
+                if j == (f1.modes-1):
+                    n -= current_state[j+1]
+                    m += 1
+                elif j<(f1.modes-1):
+                    m -= current_state[j+1] - 1
+                else:
+                    n -= current_state[j+1] - 1
+                current_state = current_state[:j] + (current_state[j]+1, 0) + current_state[j+2:]
+           
+        
+    def ngenerator(self, k, N1, N2, q):
+        local_exc2 = np.delete(self.local_exc1, k)
+        current_state = (0,)*len(local_exc2)
+        n = 0
+        m = 0
+        j = len(self.local_exc1)
+        while True:
+            pp = list(current_state)
+            pp.insert(k,q)
+            pp = tuple(pp)
+            
+            yield self.find_index[pp]
+            
+            j = len(local_exc2) - 1
+            current_state = current_state[:j] + (current_state[j]+1,)
+            n += 1
+            while n > N1 or m>N2 or current_state[j] >= local_exc2[j]:
+                j -= 1
+                if j < 0:
+                    return
+                if j == (f1.modes-1):
+                    n -= current_state[j+1]
+                    m += 1
+                elif j<(f1.modes-1):
+                    m -= current_state[j+1] - 1
+                else:
+                    n -= current_state[j+1] - 1
+                current_state = current_state[:j] + (current_state[j]+1, 0) + current_state[j+2:]
+                
+    def ptrace(self, psi, k):
+        psi1 = coo_array(psi)
+        rho = psi1.conj().T @ psi1
+        
+        N = int(self.local_exc[k])
+        N1 = f1.global_exc
+        N2 = f2.global_exc
+        border = f1.modes
+        
+        result = np.zeros((N+1,N+1), dtype = complex)
+        for p in range(N+1):
+            for q in range(N+1):
+                if k < border:
+                    N1 -= max(p,q)
+                else:
+                    N2 -= max(p,q)
+                a = list(self.ngenerator(k, N1, N2, p))
+                b = list(self.ngenerator(k, N1, N2, q))
+                for i in range(len(a)):
+                    result[p,q] += rho[a[i],b[i]]
+        return(result)
+                    
+                    
+                    
+                    
+    
     def sigmax(self, i):
         if (i<self.f1.modes):
              return(sparse.kron(self.f1.sigmax(i), self.f2.eye))
@@ -317,8 +448,9 @@ class fock_space_kron:
             s = list(state)
             return(self.f1.index(state[:self.f1.modes]) * self.f2.dimension + self.f2.index(state[self.f1.modes:]))
         
-    def real_time_solver(psi0, dt, tmax, H, Q = None, final_state = None, H_int = None):
-  
+        
+        
+def real_time_solver(psi0, dt, tmax, H, Q = None, final_state = None):
     K = psi0.size
     Nt = int(tmax/dt)+1
     psi = [psi0, psi0]
@@ -327,118 +459,67 @@ class fock_space_kron:
         H_const = H
         def H(t):
             return(H_const)
+     
+    if Q == None:
+        for i in range(1,Nt):
+            psi_iter_old = psi0
+            psi_iter = np.zeros(K)
+            psi_compare = psi0
+            while (LA.norm(psi_iter-psi_compare)>10**(-6)):
+                s = psi[1]+psi_iter_old
+                psi_iter = psi[1]+ dt/2*(-1j) * H(i*dt+dt/2)@s
+                psi_compare = psi_iter_old
+                psi_iter_old = psi_iter
+            psi[1] = psi_iter
+            
+        return (psi[1])
         
+    elif type(Q) == list:
+        l = len(Q)
+        results= np.zeros((l,Nt))
+        for j in range(l):
+            results[j,0] = abs(np.conj(psi0) @ Q[j] @ psi0)
         
-    if H_int = None:
-        
-        if Q == None:
-            for i in range(1,Nt):
-                psi_iter_old = psi0
-                psi_iter = np.zeros(K)
-                psi_compare = psi0
-                while (LA.norm(psi_iter-psi_compare)>10**(-6)):
-                    s = psi[1]+psi_iter_old
-                    psi_iter = psi[1]+ dt/2*(-1j) * H(i*dt+dt/2)@s
-                    psi_compare = psi_iter_old
-                    psi_iter_old = psi_iter
-                psi[1] = psi_iter
-                
-            return (psi[1])
+        for i in range(1,Nt):
+            psi_iter_old = psi0
+            psi_iter = np.zeros(K)
+            psi_compare = psi0
+            while (LA.norm(psi_iter-psi_compare)>10**(-6)):
+                s = psi[1]+psi_iter_old
+                psi_iter = psi[1]+ dt/2*(-1j) * H(i*dt+dt/2)@s
+                psi_compare = psi_iter_old
+                psi_iter_old = psi_iter
+            psi[1] = psi_iter
             
-        elif type(Q) == list:
-            l = len(Q)
-            results= np.zeros((l,Nt))
             for j in range(l):
-                results[j,0] = abs(np.conj(psi0) @ Q[j] @ psi0)
-            
-            for i in range(1,Nt):
-                psi_iter_old = psi0
-                psi_iter = np.zeros(K)
-                psi_compare = psi0
-                while (LA.norm(psi_iter-psi_compare)>10**(-6)):
-                    s = psi[1]+psi_iter_old
-                    psi_iter = psi[1]+ dt/2*(-1j) * H(i*dt+dt/2)@s
-                    psi_compare = psi_iter_old
-                    psi_iter_old = psi_iter
-                psi[1] = psi_iter
-                
-                for j in range(l):
-                    results[j,i] = abs(np.conj(psi[1]) @ Q[j] @ psi[1])
-            
-        else:
-            results = np.zeros(Nt)
-            results[0] = abs(np.conj(psi0) @ Q @ psi0)
-            
-            for i in range(1,Nt):
-                psi_iter_old = psi0
-                psi_iter = np.zeros(K)
-                psi_compare = psi0
-                while (LA.norm(psi_iter-psi_compare)>10**(-6)):
-                    s = psi[1]+psi_iter_old
-                    psi_iter = psi[1]+ dt/2*(-1j) * H(i*dt+dt/2)@s
-                    psi_compare = psi_iter_old
-                    psi_iter_old = psi_iter
-                psi[1] = psi_iter
-                results[i] = abs(np.conj(psi[1]) @ Q @ psi[1])
-        if final_state != None :
-            return (results, psi[1])
-        else:
-            return(results)
+                results[j,i] = abs(np.conj(psi[1]) @ Q[j] @ psi[1])
         
     else:
-        if Q == None:
-            for i in range(1,Nt):
-                psi_iter_old = psi0
-                psi_iter = np.zeros(K)
-                psi_compare = psi0
-                while (LA.norm(psi_iter-psi_compare)>10**(-6)):
-                    s = psi[1]+psi_iter_old
-                    psi_iter = psi[1]+ dt/2*(-1j) * H(i+1/2)@s
-                    psi_compare = psi_iter_old
-                    psi_iter_old = psi_iter
-                psi[1] = psi_iter
-                
-            return (psi[1])
-            
-        elif type(Q) == list:
-            l = len(Q)
-            results= np.zeros((l,Nt))
-            for j in range(l):
-                results[j,0] = abs(np.conj(psi0) @ Q[j] @ psi0)
-            
-            for i in range(1,Nt):
-                psi_iter_old = psi0
-                psi_iter = np.zeros(K)
-                psi_compare = psi0
-                while (LA.norm(psi_iter-psi_compare)>10**(-6)):
-                    s = psi[1]+psi_iter_old
-                    psi_iter = psi[1]+ dt/2*(-1j) * H(i+1/2)@s
-                    psi_compare = psi_iter_old
-                    psi_iter_old = psi_iter
-                psi[1] = psi_iter
-                
-                for j in range(l):
-                    results[j,i] = abs(np.conj(psi[1]) @ Q[j] @ psi[1])
-            
-        else:
-            results = np.zeros(Nt)
-            results[0] = abs(np.conj(psi0) @ Q @ psi0)
-            
-            for i in range(1,Nt):
-                psi_iter_old = psi0
-                psi_iter = np.zeros(K)
-                psi_compare = psi0
-                while (LA.norm(psi_iter-psi_compare)>10**(-6)):
-                    s = psi[1]+psi_iter_old
-                    psi_iter = psi[1]+ dt/2*(-1j) * H(i+1/2)@s
-                    psi_compare = psi_iter_old
-                    psi_iter_old = psi_iter
-                psi[1] = psi_iter
-                results[i] = abs(np.conj(psi[1]) @ Q @ psi[1])
-        if final_state != None :
-            return (results, psi[1])
-        else:
-            return(results)
+        results = np.zeros(Nt)
+        results[0] = abs(np.conj(psi0) @ Q @ psi0)
+        
+        for i in range(1,Nt):
+            psi_iter_old = psi0
+            psi_iter = np.zeros(K)
+            psi_compare = psi0
+            while (LA.norm(psi_iter-psi_compare)>10**(-6)):
+                s = psi[1]+psi_iter_old
+                psi_iter = psi[1]+ dt/2*(-1j) * H(i*dt+dt/2)@s
+                psi_compare = psi_iter_old
+                psi_iter_old = psi_iter
+            psi[1] = psi_iter
+            results[i] = abs(np.conj(psi[1]) @ Q @ psi[1])
+    if final_state != None :
+        return (results, psi[1])
+    else:
+        return(results)
+    
         
         
 
+
+# In[ ]:
+
+
+
+
